@@ -209,7 +209,7 @@ class ReproduceRewards(nn.Module):
         reward_seq = []
         for act, next_ob in zip(acts, next_obs):
             _act, _ = self.policy.reproduce(init_ob, act)
-            _next_ob, _ = self.model.reproduce(init_ob, _act, next_ob)
+            _next_ob, _ = self.model.reproduce(next_ob, self.model(init_ob, _act))
             reward_seq.append(self.reward_fn(init_ob, _act, _next_ob))
             init_ob = _next_ob
         return torch.stack(reward_seq)
